/*
 * SBPool.java
 *
 * Created on January 21, 2003, 9:27 AM
 */

package com.divnull.util;

/**
 * A singleton wrapper class for a {@link StringBufferPool}. Most most 
 * purposes, using the routines in this class is preferable to using the 
 * pool directly.
 * <p>
 * Like all pools, you are responsible for returning objects that you
 * borrow. Here is an example that gets a 64 character buffer:
 * <blockquote><code>StringBuffer buffer = SBPool.borrowBuffer64();</code>
 * <br>//... do your thing with the buffer ...
 * <br><code>String str = SBPool.returnBufferGetString(buffer);</code></blockquote>
 * As with string buffers in general, you should get the largest buffer
 * that you think you need. Try to avoid exceeding the buffer size, as this
 * makes the buffer grow and slows performance. This example gets a 
 * 256 character buffer:
 * <blockquote><code>StringBuffer buffer = SBPool.borrowBuffer256();
 * <br></code>//... do your thing with the buffer ...
 * <br><code>String str = SBPool.returnBuffer(buffer);</code></blockquote>
 * If you are getting a string buffer in the middle of a try block, 
 * make sure to clean up in the <code>finally</code> statement, like
 * so:
 * <blockquote><code>try
 * <br>{
 * <br>&nbsp;&nbsp;&nbsp;StringBuffer buffer = SBPool.borrowBuffer64();
 * <br>&nbsp;&nbsp;&nbsp;</code>//... do you thing with the buffer ...<code>
 * <br>}
 * <br>catch (Exception e)
 * <br>{
 * <br>&nbsp;&nbsp;&nbsp;String str = SBPool.returnBuffer(buffer);
 * <br>}</code></blockquote>
 * @author  lward
 */
public class SBPool
{
    /* Singleton pool class. It is private to force subclasses to call getPool 
     * to access it. */
    private static StringBufferPool m_Pool = null;
            
    /*------- Public static methods -------*/
    
    /** Get a string buffer preallocated to an arbitrary size.
     * <p>
     * This routine should be avoided, when possible, in favor of one of the 
     * methods that return a buffer of one of the "standard" sizes. Performance 
     * of the pool is optimized when these standard sizes are used.
     * @param size The preallocated size of the buffer.
     * @return A buffer of the requested size.    
     * @see #borrowBuffer32
     * @see #borrowBuffer64
     * @see #borrowBuffer128
     * @see #borrowBuffer256
     * @see #borrowBuffer512
     * @see #borrowBuffer1k
     * @see #borrowBuffer2k
     * @see #borrowBuffer4k
     */
    public static final StringBuffer borrowBuffer( int size )
    {
        StringBuffer buffer = null;
        try
        {
            StringBufferPool pool = getPool();
            buffer = pool.borrowBuffer(size);
        }
        catch (Exception e)
        {
            Integer bufSize = new Integer(size);
            buffer = handleExceptionOnBorrow( e,bufSize );
        }
        return buffer;
    }
    
    /** Get a string buffer preallocated to 32 characters.
     * @return A 32 character buffer */    
    public static final StringBuffer borrowBuffer32()
    {
        StringBuffer buffer = null;
        try
        {
            StringBufferPool pool = getPool();
            buffer = pool.borrowBuffer32();
        }
        catch (Exception e)
        {
            buffer = handleExceptionOnBorrow(e,StringBufferPool.kMicroStringSize);
        }
        return buffer;
    }
 
    /** Get a string buffer preallocated to 64 characters.
     * @return A 64 character buffer */    
    public static final StringBuffer borrowBuffer64()
    {
        StringBuffer buffer = null;
        try
        {
            StringBufferPool pool = getPool();
            buffer = pool.borrowBuffer64();
        }
        catch (Exception e)
        {
            buffer = handleExceptionOnBorrow(e,StringBufferPool.kTinyStringSize);
        }
        return buffer;
    }
 
    /** Get a string buffer preallocated to 128 characters.
     * @return A 128 character buffer */    
    public static final StringBuffer borrowBuffer128()
    {
        StringBuffer buffer = null;
        try
        {
            StringBufferPool pool = getPool();
            buffer = pool.borrowBuffer128();
        }
        catch (Exception e)
        {
            buffer = handleExceptionOnBorrow(e,StringBufferPool.kSmallStringSize);
        }
        return buffer;
    }
 
    /** Get a string buffer preallocated to 256 characters.
     * @return A 256 character buffer */    
    public static final StringBuffer borrowBuffer256()
    {
        StringBuffer buffer = null;
        try
        {
            StringBufferPool pool = getPool();
            buffer = pool.borrowBuffer256();
        }
        catch (Exception e)
        {
            buffer = handleExceptionOnBorrow(e,StringBufferPool.kMediumStringSize);
        }
        return buffer;
    }
 
    /** Get a string buffer preallocated to 512 characters.
     * @return A 512 character buffer */    
    public static final StringBuffer borrowBuffer512()
    {
        StringBuffer buffer = null;
        try
        {
            StringBufferPool pool = getPool();
            buffer = pool.borrowBuffer512();
        }
        catch (Exception e)
        {
            buffer = handleExceptionOnBorrow(e,StringBufferPool.kLargeStringSize);
        }
        return buffer;
    }
 
    /** Get a string buffer preallocated to 1024 characters.
     * @return A 1024 character buffer */    
    public static final StringBuffer borrowBuffer1k()
    {
        StringBuffer buffer = null;
        try
        {
            StringBufferPool pool = getPool();
            buffer = pool.borrowBuffer1k();
        }
        catch (Exception e)
        {
            buffer = handleExceptionOnBorrow(e,StringBufferPool.kHugeStringSize);
        }
        return buffer;
    }
 
    /** Get a string buffer preallocated to 2048 characters.
     * @return A 2048 character buffer */    
    public static final StringBuffer borrowBuffer2k()
    {
        StringBuffer buffer = null;
        try
        {
            StringBufferPool pool = getPool();
            buffer = pool.borrowBuffer2k();
        }
        catch (Exception e)
        {
            buffer = handleExceptionOnBorrow(e,StringBufferPool.kGiantStringSize);
        }
        return buffer;
    }
 
    /** Get a string buffer preallocated to 4096 characters.
     * @return A 4096 character buffer */    
    public static final StringBuffer borrowBuffer4k()
    {
        StringBuffer buffer = null;
        try
        {
            StringBufferPool pool = getPool();
            buffer = pool.borrowBuffer4k();
        }
        catch (Exception e)
        {
            buffer = handleExceptionOnBorrow(e,StringBufferPool.kGargantuanStringSize);
        }
        return buffer;
    }
 
    /** Return a string buffer to the pool.
     * @param buffer A buffer previously borrowed from the pool.
     * @see #borrowBuffer32
     * @see #borrowBuffer64
     * @see #borrowBuffer128
     * @see #borrowBuffer256
     * @see #borrowBuffer512
     * @see #borrowBuffer1k
     * @see #borrowBuffer2k
     * @see #borrowBuffer4k
     */    
    public static final void returnBuffer(StringBuffer buffer)
    {
        try
        {
            StringBufferPool pool = getPool();
            pool.returnBuffer( buffer );
        }
        catch (Exception e)
        {
            handleExceptionOnReturn( e,buffer );
        }
    }    
    
    /** Return a string buffer to the pool and get back the string object
     * from the buffer.
     * @param buffer A buffer previously borrowed from the pool.
     * @return The string object rendered from the buffer.
     * @see #borrowBuffer32
     * @see #borrowBuffer64
     * @see #borrowBuffer128
     * @see #borrowBuffer256
     * @see #borrowBuffer512
     * @see #borrowBuffer1k
     * @see #borrowBuffer2k
     * @see #borrowBuffer4k
     */    
    public static final String returnBufferGetString(StringBuffer buffer)
    {
        String str = null;
        try
        {
            StringBufferPool pool = getPool();
            str = pool.returnBufferGetString(buffer);
        }
        catch (Exception e)
        {
            handleExceptionOnReturn( e,buffer );
        }
        return str;
    }
    
    
    /** Prepopulate the pool with a given number of string buffers of a given
     * size.
     * @param size The size to allocate the string buffers. This will be pinned
     * to a standard size, if possible.
     * @param count The number of buffers of that size to create. */
    public void prePopulatePool( int size,int count ) throws Exception
    {
        try
        {
            StringBufferPool pool = getPool();
            pool.prePopulatePool(size,count);
        }
        catch (Exception e)
        {
            handleExceptionOnPrepopulate( e,size,count );
        }
    }
    
    
   /** Gets the number of idle string buffers currently pooled under a given key.
    * @param key The key to look for.
    * @return int The number of idle strings for that key.
    */
    public static int getNumIdle(Integer key)
    {
        StringBufferPool pool = getPool();
        return pool.getNumIdle(key);
    }

    
    /** An internal accessor to the pool singleton. This creates the singleton
     * if it has not been created.
     * @return The string buffer pool.
     */    
    protected static StringBufferPool getPool()
    {
        if ( m_Pool == null )
        {
            m_Pool = new StringBufferPool();
            StringBufferPoolableObjectFactory factory = new StringBufferPoolableObjectFactory();
            m_Pool.setFactory(factory);
        }
        return m_Pool;
    }
    
    
    /** Handles an exception thrown during an attempt to borrow a string buffer 
     * from the pool.
     * <p>
     * This implementation ignores the exception and returns a buffer from 
     * outside the pool. Note attempting to return such a buffer to the pool 
     * should, in fact, succeed, assuming the pool in general is in good shape
     * (which, given that an exception was thrown, it may not be).
     * @param e The exception that was thrown
     * @param size The size of the string buffer requested.
     * @return A string buffer from outside the pool.
     */
    protected static StringBuffer handleExceptionOnBorrow( Exception e,Integer size )
    {
        StringBuffer buffer = new StringBuffer( size.intValue() );
        return buffer;
    }
    
    /** Handles an exception thrown during an attempt to return a string buffer
     * to the pool.
     * <p>
     * This implementation ignores the exception.
     * @param e The exception that was thrown
     * @param buffer The buffer being returned.
     */
    protected static void handleExceptionOnReturn( Exception e,StringBuffer buffer )
    {
        /* Do nothing */
    }


    /** Handles an exception thrown during an attempt to preallocate string 
     * buffers in the pool.
     * <p>
     * This implementation ignores the exception.
     * @param e The exception that was thrown
     * @param size The size being preallocated.
     * @param size The number of buffers of the given key to preallocate.
     */
    protected static void handleExceptionOnPrepopulate( Exception e,int size,int count )
    {
        /* Do nothing */
    }


    /** Creates a new instance of SBPool */
    public SBPool()
    {
    }
}
